import json
from enum import Enum
from http import HTTPStatus
from dataclasses import dataclass
from typing import Generic, TypeVar, Type
from pydantic import BaseModel, Field

from ..utils.convert import try_str_to_float
from ..utils.data import Order, OrderType, Condition, ConditionType, Pageable

T = TypeVar("T")


####################################
# Cognito
####################################

class EventType(Enum):
    sqs = 1
    cognito = 2


class CallerContext(BaseModel):
    aws_sdk_version: str = Field(alias="awsSdkVersion")
    client_id: str = Field(alias="clientId")


class UserPoolEvent(BaseModel):
    version: str
    trigger_source: str = Field(alias="triggerSource")
    region: str
    user_pool_id: str = Field(alias="userPoolId")
    user_name: str = Field(alias="userName")
    caller_context: CallerContext = Field(alias="callerContext")
    request: dict
    response: dict


####################################
# Rest API
####################################

reserved_keys = ["fields", "pageNo", "pageSize", "topSize", "order"]


@dataclass()
class HTTPResponse:
    status: HTTPStatus = HTTPStatus.OK
    body: any = None
    headers: dict[str, any] = None


@dataclass()
class QueryParams:
    fields: list[str] = None
    conditions: list[Condition] = None
    page: Pageable = Pageable()
    order: list[Order] = None


@dataclass()
class AuthUser:
    cognito_id: str = None
    email: str = None
    email_confirmed: bool = False


@dataclass()
class HTTPRequest(Generic[T]):
    body: T = None
    headers: dict[str, any] = None
    path_params: dict[str, any] = None
    user: AuthUser = AuthUser()
    query_params: QueryParams = QueryParams()


def http_event_to_request(cls: Type[T], event: dict[str, any], context) -> HTTPRequest[T]:
    data = json.loads(event["body"]) if event.get("body") else None

    request = HTTPRequest(body=cls(**data) if cls is not dict else data,
                          headers=event.get("headers"),
                          path_params=event.get("pathParameters"))

    if "authorizer" in event["requestContext"]:
        claims = event["requestContext"]["authorizer"]["claims"]
        request.user = AuthUser(claims["sub"], claims["email"], claims["email_verified"] == "true")

    params: dict[str, any] | None = event.get("queryStringParameters")

    if params:
        fields = params["fields"].replace(" ", "").split(",") if params.get("fields") else list()
        pageable = Pageable(page_param(params, alias="pageNo"),
                            page_param(params, alias="pageSize", default=20, max_=50, min_=1),
                            page_param(params, alias="topSize", max_=1000, min_=-1))

        conditions = list()
        for k, v in params.items():
            key = k.strip()
            if key in reserved_keys:
                continue
            else:
                value = v.strip()
                value_ = value.lower()
                if value_.startswith("range(") and value_.endswith(")"):
                    condition_type = ConditionType.range
                    data = condition_param("range", value)
                    if not len(data) >= 2:
                        continue
                elif value_.startswith("any(") and value_.endswith(")"):
                    condition_type = ConditionType.any
                    data = condition_param("any", value)
                else:
                    condition_type = ConditionType.compare
                    data = try_str_to_float(value)
                conditions.append(Condition(condition_type, key, data))

        order = list()
        for order_ in (params["order"].split(",") if params.get("order") else list()):
            o = order_.strip().split(" ")
            if len(o) < 1:
                continue
            order_type = OrderType.asc
            if len(o) == 2:
                order_type = next((e for e in OrderType if e.value == o[1].upper()), None)
            if not order_type:
                continue
            order.append(Order(order_type, alias=o[0]))

        request.query_params = QueryParams(fields, conditions, pageable, order)

    return request


def page_param(params: dict[str, any], alias: str, default: int = 0, max_: int = None, min_: int = 0) -> int:
    try:
        if not params.get(alias):
            return default
        value_ = int(params[alias])
        return value_ if (max_ >= value_ >= min_ if max_ else value_ >= min_) else default
    except ValueError:
        return default


def condition_param(type_, p) -> list:
    return [try_str_to_float(i) for i in p[len(type_) + 1:len(p) - 1].replace(" ", "").split(",")]
